@@ -680,8 +680,8 @@ static void ifx_xhomezr_notify(GAtResult *result, gpointer user_data)
 
 static void ifx_xciev_notify(GAtResult *result, gpointer user_data)
 {
-	struct ofono_netreg *netreg = user_data;
-	int strength, ind;
+	//struct ofono_netreg *netreg = user_data;
+	int ind;
 	GAtResultIter iter;
 
 	g_at_result_iter_init(&iter, result);
@@ -692,12 +692,39 @@ static void ifx_xciev_notify(GAtResult *result, gpointer user_data)
 	if (!g_at_result_iter_next_number(&iter, &ind))
 		return;
 
-	if (ind == 0)
+	DBG("ind %d", ind);
+
+	/*
+	 * Radio signal strength indicators are defined for 0-7,
+	 * but this notification seems to return CSQ 0-31,99 values.
+	 *
+	 * Ignore this indication for now since it can not be trusted.
+	 */
+}
+
+static void ifx_xcsq_notify(GAtResult *result, gpointer user_data)
+{
+	struct ofono_netreg *netreg = user_data;
+	int rssi, ber, strength;
+	GAtResultIter iter;
+
+	g_at_result_iter_init(&iter, result);
+
+	if (!g_at_result_iter_next(&iter, "+XCSQ:"))
+		return;
+
+	if (!g_at_result_iter_next_number(&iter, &rssi))
+		return;
+
+	if (!g_at_result_iter_next_number(&iter, &ber))
+		return;
+
+	DBG("rssi %d ber %d", rssi, ber);
+
+	if (rssi == 99)
 		strength = -1;
-	else if (ind == 7)
-		strength = 100;
 	else
-		strength = (ind * 15);
+		strength = (rssi * 100) / 31;
 
 	ofono_netreg_strength_notify(netreg, strength);
 }
@@ -1470,6 +1497,10 @@ static void at_creg_set_cb(gboolean ok, GAtResult *result, gpointer user_data)
 		/* Register for specific signal strength reports */
 		g_at_chat_register(nd->chat, "+XCIEV:", ifx_xciev_notify,
 						FALSE, netreg, NULL);
+		g_at_chat_register(nd->chat, "+XCSQ:", ifx_xcsq_notify,
+						FALSE, netreg, NULL);
+		g_at_chat_send(nd->chat, "AT+XCSQ=1", none_prefix,
+						NULL, NULL, NULL);
 		g_at_chat_send(nd->chat, "AT+XMER=1", none_prefix,
 						NULL, NULL, NULL);
 
